﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Evt = EventStore.ClientAPI;
using System.Timers;  

namespace OrphanConnectionTest
{
  partial class EventStore
  {
    #region Persistent subscription methods
    private void OnSubscriptionDropped(string stream, string groupName, Evt.SubscriptionDropReason dropReason, Exception exception)
    {
      try
      {
        SubscriptionDetails details = null;
        if (_subscriptions.TryRemove(GetSubscriptionKey(stream, groupName), out details))
        {
          _droppedSubscriptions.Enqueue(details);
          _subscriptionReconnectCheckTimer?.Start();
          Logger.Log($"{stream} - {groupName}- Subscription dropped and has been queued.");
         
        }
        else
        {
          Logger.Log($"{stream} - {groupName}- Subscription dropped and is not found in internal dictionary.");
        }
      }
      catch (Exception e)
      {
        Logger.Log(nameof(OnSubscriptionDropped), e);
      }

    }
    #endregion

    #region Connection methods
    private void OnAuthenticationFailed(object sender, Evt.ClientAuthenticationFailedEventArgs args)
    {
      Logger.Log("Connection authentication failed " + args.Reason);
    }
    private void OnErrorOccured(object sender, Evt.ClientErrorEventArgs args)
    {
      Logger.Log("Error occurred in connection " + args.Exception.ToString());
    }

    private void OnClosed(object sender, Evt.ClientClosedEventArgs args)
    {
      try
      {
        Logger.Log("Connection has been closed.");
      }
      catch (Exception e)
      {
        Logger.Log(nameof(OnClosed), e);
      }

    }

    private async void OnConnected(object sender, Evt.ClientConnectionEventArgs args)
    {
      await ConnectToSubscription();
      _reconnectCount = 0;
    }

    private void OnDisconnected(object sender, Evt.ClientConnectionEventArgs args)
    {
      try
      {
        Logger.Log("Connection has been disconnected.");
      }
      catch (Exception e)
      {
        Logger.Log(nameof(OnDisconnected), e);
      }
    }

    private void OnReconnecting(object sender, Evt.ClientReconnectingEventArgs args)
    {
      if (_reconnectCount == 0 || _reconnectCount > RECONNECT_LOG_COUNT)
      {
        Logger.Log("Attempting to reconnect.");
        _reconnectCount = 1;
      }
      else
        _reconnectCount++; //Avoid logging frequently         
    }

    #endregion

    #region Helper methods
    private async void OnSubscriptionReconnectCheckTimer_Elapsed(object sender, ElapsedEventArgs e)
    {
      try
      {
        Logger.Log("Subscription reconnect check timer elapsed.");
        await ConnectToSubscription();
      }
      catch (Exception ex)
      {
        Logger.Log(nameof(OnSubscriptionReconnectCheckTimer_Elapsed), ex);
      }
    }


    private async Task ConnectToSubscription()
    {
      if (!CanConnect())
      {
        Logger.Log("Another connection in process");
        return;
      }

      try
      {
        if (_subscriptionReconnectCheckTimer.Enabled)
          _subscriptionReconnectCheckTimer.Stop();

        Logger.Log("Connecting");
        bool fRaised = false;
        SubscriptionDetails details;
        while (_droppedSubscriptions.TryDequeue(out details))
        {
          if (!fRaised)
          {           
           
            fRaised = true;
          }
          try
          {
            await ConnectToPersistentSubscription(details.StreamName,details.GroupName);
          }
          catch (Exception)
          {
            _droppedSubscriptions.Enqueue(details);
            throw;
          }
          Logger.Log($"{details.StreamName} - {details.GroupName} - subscribed on connect.");
        }
      }
      catch (Exception e)
      {
        Logger.Log(nameof(ConnectToSubscription), e);
      }
      finally
      {
        FinishConnecting();
        if (!_subscriptionReconnectCheckTimer.Enabled)
          _subscriptionReconnectCheckTimer.Start();
      }
    }

    private bool CanConnect()
    {
      lock (_connectLock)
      {
        if (_connecting) //If already connecting return;
          return false;
        else
        {
          _connecting = true;
          return true;
        }
      }
    }

    private void FinishConnecting()
    {
      lock (_connectLock)
      {
        _connecting = false;
      }
    }
    #endregion

  }
}

