﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Configuration;
using Evt = EventStore.ClientAPI;
using System.Collections.Concurrent;
using System.Timers;

namespace OrphanConnectionTest
{
  partial class EventStore: IDisposable
  {
    private Uri _tcpUrl;
    private Uri _httpUrl;
    private string _user;
    private string _pswd;
    private Evt.IEventStoreConnection _connection;    
    private bool _connecting;
    private byte[] _connectLock;
    private ConcurrentDictionary<string, SubscriptionDetails> _subscriptions;
    private ConcurrentQueue<SubscriptionDetails> _droppedSubscriptions;
    private const int RECONNECT_LOG_COUNT = 300;
    private static int _reconnectCount = 0;
    private Timer _subscriptionReconnectCheckTimer;
    private const double SUBSCRIPTION_RECONNECT_CHECK_TIME = 300000;
    private ConcurrentDictionary<string, bool> _subscriptionConnectionLock;

    public EventStore()
    {
      _subscriptions = new ConcurrentDictionary<string, SubscriptionDetails>();
      _droppedSubscriptions = new ConcurrentQueue<SubscriptionDetails>();
      _tcpUrl = new Uri(ConfigurationManager.AppSettings["tcp"]);
      _httpUrl = new Uri(ConfigurationManager.AppSettings["http"]);
      _user = ConfigurationManager.AppSettings["user"];
      _pswd = ConfigurationManager.AppSettings["password"];
      _connecting = false;
      _connectLock = new byte[0];
      _subscriptionReconnectCheckTimer = new Timer(SUBSCRIPTION_RECONNECT_CHECK_TIME);
      _subscriptionReconnectCheckTimer.Elapsed += OnSubscriptionReconnectCheckTimer_Elapsed;
      _subscriptionReconnectCheckTimer.Start();
      _subscriptionConnectionLock = new ConcurrentDictionary<string, bool>();
    }

    private static string GetSubscriptionKey(string streamName, string groupName)
    {
      return $"{streamName}/{groupName}";
    }

    public void Disconnect()
    {
      _subscriptionConnectionLock.Clear();
      if (_connection != null)
      {
        try
        {
          _connection.AuthenticationFailed -= OnAuthenticationFailed;
          _connection.Closed -= OnClosed;
          _connection.Connected -= OnConnected;
          _connection.Disconnected -= OnDisconnected;
          _connection.ErrorOccurred -= OnErrorOccured;
          _connection.Reconnecting -= OnReconnecting;
          _connection.Close();
        }
        catch { }
        try
        {
          _connection.Dispose();
        }
        catch { }
        _connection = null;
      }

      try
      {
        if (_subscriptionReconnectCheckTimer != null)
        {
          _subscriptionReconnectCheckTimer.Elapsed -= OnSubscriptionReconnectCheckTimer_Elapsed;
          _subscriptionReconnectCheckTimer.Dispose();
          _subscriptionReconnectCheckTimer = null;
        }
      }
      catch { }
      Logger.Log("Connection closed");
    }

    public async Task Connect()
    {

      var evtSettings = Evt.ConnectionSettings.Create().
        KeepReconnecting().
        KeepRetrying().
        SetHeartbeatTimeout(new TimeSpan(0, 0, 60)).
        SetHeartbeatInterval(new TimeSpan(0, 0, 30)).
        SetDefaultUserCredentials(new Evt.SystemData.UserCredentials(_user, _pswd)).
        Build();
      _connection = Evt.EventStoreConnection.Create(evtSettings,
                                                    _tcpUrl,
                                                    _tcpUrl.OriginalString);

      _connection.AuthenticationFailed += OnAuthenticationFailed;
      _connection.Closed += OnClosed;
      _connection.Connected += OnConnected;
      _connection.Disconnected += OnDisconnected;
      _connection.ErrorOccurred += OnErrorOccured;
      _connection.Reconnecting += OnReconnecting;
      await _connection.ConnectAsync();
      Logger.Log("Connection opened");
    }

    public async Task UpsertAndSubscribeToPersistentSubscriptionAsync(string streamName, string groupName)
    {
      
      SubscriptionDetails value;
      if (_subscriptions.TryGetValue(GetSubscriptionKey(streamName, groupName), out value))
      {
        Logger.Log($"{GetSubscriptionKey(streamName, groupName)} is already subscribed.");
      }
      else
      {        
        await CreateOrUpdatePersistentSubscriptionAsync(streamName, groupName);
        await ConnectToPersistentSubscription(streamName,groupName);
      }

    }

    private async Task ConnectToPersistentSubscription(string streamName, string groupName)
    {
      //Do not allow more than one connection to the subscription group for the same connection. 
      //Eventstore opens multiple subscriber connections and causes weird behavior
      if (_subscriptionConnectionLock.TryAdd(GetSubscriptionKey(streamName, groupName), true))
      {
        SubscriptionDetails details;
        if (_subscriptions.TryGetValue(GetSubscriptionKey(streamName, groupName), out details))
        {
          Logger.Log($"{GetSubscriptionKey(streamName, groupName)} is already subscribed.");
        }
        else
        {
          try
          {
            var result = await _connection.ConnectToPersistentSubscriptionAsync(
                             streamName,
                             groupName,
                             (subscription, resolved) => Task.Run(() => OnEventAppearedPersistentStream(subscription, GetSubscriptionKey(streamName, groupName), resolved, false)),
                             (subscription, dropReason, exception) => Task.Run(() => OnSubscriptionDropped(streamName, groupName, dropReason, exception)),
                             null,
                             Environment.ProcessorCount,
                             false);
            _subscriptions.TryAdd(GetSubscriptionKey(streamName, groupName), new SubscriptionDetails(result, streamName, groupName));
            Logger.Log($"{GetSubscriptionKey(streamName,groupName)} subscribed.");
          }
          finally
          {
            bool value;
            _subscriptionConnectionLock.TryRemove(GetSubscriptionKey(streamName, groupName), out value);
          }
        }
      }
      return;
    }

    private void OnEventAppearedPersistentStream(
           Evt.EventStorePersistentSubscriptionBase subscription,
           string subscriptionKey,
           Evt.ResolvedEvent @event,
           bool autoAck)
    {
      try
      {
        Logger.Log($"{@event.Event.EventId} - {subscriptionKey} - Received");
        subscription.Acknowledge(@event);
        Logger.Log($"{@event.Event.EventId} - {subscriptionKey} - Acknowledged");
      }
      catch (Exception ex)
      {
        Logger.Log($"{@event.Event.EventId} - {subscriptionKey} - {autoAck.ToString()} unsubscribed.", ex);
      }
    }

    private async Task CreateOrUpdatePersistentSubscriptionAsync(string streamName, string groupName)
    {
      var evtSettings = Evt.PersistentSubscriptionSettings
                      .Create()
                      .WithMessageTimeoutOf(new TimeSpan(0, 0, 0, 0, 10000))
                      .WithMaxRetriesOf(Int32.MaxValue)
                      .CheckPointAfter(new TimeSpan(0, 0, 0, 31))
                      .MaximumCheckPointCountOf(30)
                      .MinimumCheckPointCountOf(1)
                      .WithMaxSubscriberCountOf(100)
                      .Build();


      try
      {
        // TODO: Trace return status for troubleshooting. This method may fail if the group already exists. 
        // EventStore.ClientAPI does not return this value due to a defect
        // https://github.com/EventStore/EventStore/issues/1053

        await _connection.CreatePersistentSubscriptionAsync(streamName,
                                                      groupName,
                                                      evtSettings,
                                                      null);
        Logger.Log($"{GetSubscriptionKey(streamName, groupName)} created.");
      }
      catch (InvalidOperationException e)
      {
        Logger.Log($"{GetSubscriptionKey(streamName, groupName)} already exists ({e.Message})");

      }
      catch (Exception e)
      {
        Logger.Log($"{GetSubscriptionKey(streamName, groupName)}", e);
        //throw e;
      }

      return;
    }

    public void Dispose()
    {
      Disconnect();
    }
  }
}

